from sqlalchemy.orm import declarative_base

SQLALCHEMY_DATABASE_URL = None
DB_ENGINE = None
DBSessionLocal = None

EntityBase = declarative_base()


def get_db():
    db = DBSessionLocal()
    try:
        yield db
    finally:
        db.close()



from convergence.convergence_service import ConvergenceService


def convergence_endpoint(router, url, method='GET', exposed_through_gateway=True, authorization=None):
    ConvergenceService.register_endpoint_info(url, method, authorization, exposed_through_gateway)

    def wrapper(func):
        if method.upper() == 'GET':
            router.get(url)(func)
        elif method.upper() == 'POST':
            router.post(url)(func)
        elif method.upper() == 'PUT':
            router.put(url)(func)
        elif method.upper() == 'PATCH':
            router.patch(url)(func)
        elif method.upper() == 'DELETE':
            router.delete(url)(func)

        return func

    return wrapper



from sqlalchemy import Column, UUID, String, TIMESTAMP

from convergence.database_orm import EntityBase
from convergence.db_migrations.blueprint_formatter import SqlDialectFormatter
from convergence.db_migrations.migrations import BaseDatabaseMigration
from convergence.db_migrations.structures import TableBlueprint, TableColumnBlueprint


class DatabaseMigration(EntityBase):
    __tablename__ = "database_migrations"
    uuid = Column(UUID, primary_key=True)
    migration_name = Column(String)
    command = Column(String)
    applied_timestamp = Column(TIMESTAMP)


class BaseCreateMigrationsTable(BaseDatabaseMigration):
    def __init__(self):
        super().__init__('create_migration_table')

    def get_dependencies(self):
        return []

    def get_migration_ddl(self, formatters: SqlDialectFormatter):
        result = []

        migration_table = TableBlueprint()
        migration_table.name = "database_migrations"

        migration_table.columns.append(TableColumnBlueprint("uuid", "UUID", True, False, False, None))
        migration_table.columns.append(TableColumnBlueprint("migration_name", "String[255]", False, True, False, None))
        migration_table.columns.append(TableColumnBlueprint("command", "String", False, False, False, None))
        migration_table.columns.append(
            TableColumnBlueprint("applied_timestamp", "timestamp", False, False, False, "CURRENT_TIMESTAMP"))

        result.append(formatters.create_table_formatter.to_sql(migration_table))

        result = ''.join(result)
        return result

    def allow_failure(self):
        return False



import json

from convergence.helpers.controller_helpers import convert_object_to_dict
from convergence.internal.observability.request_log import RequestLog
from convergence.internal.observability.serializer_config import ILogSerializer


class JsonLogSerializer(ILogSerializer):
    def __init__(self, path, production):
        self.path = path
        self.production = production

    def save(self, request_log: RequestLog):
        from convergence.constants import VERSION
        # kwargs = {}
        #
        # if not self.production:
        #     kwargs['indent'] = '  '
        #
        # with open(self.path + '/' + request_log.request_identifier + '.txt', 'w') as file:
        #     json_str = jsonpickle.encode(request_log, **kwargs)
        #     file.write(json_str)
        kwargs = {}

        if not self.production:
            kwargs['indent'] = '  '

        with open(self.path + '/' + request_log.request_identifier + '.crl', 'w') as file:
            request_as_dict = convert_object_to_dict(request_log)
            request_as_dict["service_language"] = "python"
            request_as_dict["template_version"] = VERSION
            json.dump(request_as_dict, file, **kwargs)



import encodings.utf_8
import json

from starlette.middleware.base import BaseHTTPMiddleware
from fastapi import Request
from starlette.responses import JSONResponse

from convergence.dto.bad_request_details import BadRequestDetails, BadFieldRequest
from convergence.dto.base_dto import ApiResponse
from convergence.helpers import errors
from convergence.helpers.controller_helpers import convert_object_to_dict, update_response_with_request_ids_for_filters
from convergence.internal.observability.request_log import RequestLog
from convergence.internal.observability.serializer_config import LogSerializerConfig


def convert_pydantic_error_to_convergence(error_response) -> BadRequestDetails:
    result = BadRequestDetails()

    errors = {}
    for entry in error_response['detail']:
        map_id = create_map_id(entry['loc'])
        error_entry = None
        if map_id in errors:
            error_entry = errors[map_id]
        else:
            error_entry = BadFieldRequest()
            error_entry.location = entry['loc'][0]
            error_entry.field = create_map_id(entry['loc'][1:])
            result.errors.append(error_entry)
            errors[map_id] = error_entry

        error_entry.error_messages.append(entry['msg'])

    return result


def create_map_id(list_of_fields):
    result = ''

    sep = ''
    for field in list_of_fields:
        if isinstance(field, int):
            result += f'[{field}]'
        else:
            result += f'{sep}{field}'
        sep = '.'

    return result


class InputValidationMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, **kwargs):
        super().__init__(app)
        self.service = kwargs['service']
        self.log_serializer = LogSerializerConfig.create(self.service)

    async def dispatch(self, request: Request, call_next):
        response = await call_next(request)

        if response.status_code == 422:
            error_response = await self.__extract_pydantic_response(response)
            if response is None:
                response = self.__internal_error_response(request)
            elif self.__is_invalid_json(error_response):
                response = self.__invalid_json_input(request)
            else:
                response = self.__build_convergence_unified_bad_request(error_response, request)

        return response

    def __internal_error_response(self, http_request):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, http_request, self.service)

        status_code = 500

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'The request input was invalid, but failed to create a proper problem detail.'
        response.header.code = errors.INVALID_DATA
        response.header.body_type = 'failure_info'

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    async def __extract_pydantic_response(self, response):
        try:
            chunks = []
            async for chunk in response.body_iterator:
                chunk = encodings.utf_8.decode(chunk)[0]
                chunks.append(chunk)
            chunks = ''.join(chunks)
            return json.loads(chunks)
        except:
            return None

    def __build_convergence_unified_bad_request(self, error_response, http_request):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, http_request, self.service)

        status_code = 400

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'The request input is invalid, refer to body for details.'
        response.header.code = errors.INVALID_DATA
        response.body = convert_pydantic_error_to_convergence(error_response)
        response.header.body_type = response.body.get_response_body_type()

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)

    def __is_invalid_json(self, response):
        try:
            path = ['detail', 0, 'msg']
            msg = response

            for component in path:
                msg = msg[component]

            return msg is not None and msg == 'JSON decode error'
        except:  # noqa
            return False

    def __invalid_json_input(self, http_request):
        request_type = self.service.get_configuration('observability.request_id_prefix')
        request_log = RequestLog.initialize(request_type, http_request, self.service)

        status_code = 400

        response = ApiResponse()
        response.header.status_code = status_code
        response.header.message = 'The request input is not a valid JSON.'
        response.header.code = errors.API_UNPARSABLE_JSON
        response.header.body_type = "api_failure"
        response.body = None

        update_response_with_request_ids_for_filters(request_log, response, http_request)
        request_log.finish(response)
        self.log_serializer.save(request_log)

        response = convert_object_to_dict(response)
        return JSONResponse(response, status_code=status_code)



from fastapi import Request

from convergence.security.rule_expression import AuthorizationRuleExpression


class AccessControlLayer:
    def __init__(self):
        self.__rule_expression_cache = {}

    def is_authorized(self, rule: str, request: Request, token: dict):
        rule_expression = self.__parse_rule_expression(rule)
        context = {
            'acl': self,
            'req': request, 'request': request,
            'false': False, 'true': True,
            'token': token
        }

        return rule_expression.evaluate(context)

    def allow_all(self, **kwargs):
        return True

    def is_signed_in(self, **kwargs):
        token = kwargs.get('token')
        return token is not None

    def not_signed_in(self, **kwargs):
        token = kwargs.get('token')
        return token is None

    def has_authority(self, authority, **kwargs):
        token = kwargs.get('token')
        if token is None:
            return False
        authorities = token.get('authorities')
        if authorities is None:
            return False
        return authorities is not None and authority in authorities

    def is_service(self, **kwargs):
        token = kwargs.get('token')
        if token is None:
            return False
        return token is not None and 'is_inter_service_call' in token and token['is_inter_service_call']

    def is_system_admin(self, **kwargs):
        token = kwargs.get('token')
        if token is None:
            return False
        return self.has_authority(token, 'authority::*')

    def __parse_rule_expression(self, rule):
        if rule in self.__rule_expression_cache:
            return self.__rule_expression_cache[rule]

        result = AuthorizationRuleExpression()
        result.parse(rule)

        self.__rule_expression_cache[rule] = result
        return result



import abc
from typing import List

from convergence.convergence_service import ConvergenceService
from convergence.db_migrations.blueprint_formatter import SqlDialectFormatter


class DatabaseSeedSpec:
    def __init__(self):
        self.table_name = None
        self.fields = {}

    def table(self, table):
        self.table_name = table
        return self

    def field(self, name: str, value):
        self.fields[name] = value
        return self


class DatabasePreparationStep(abc.ABC):
    def __init__(self, name):
        self.name = name

    @abc.abstractmethod
    def get_dependencies(self) -> List[str]:
        pass

    def __repr__(self):
        return self.name


class BaseDatabaseSeed(DatabasePreparationStep):
    def __init__(self, name):
        super().__init__(name)

    @abc.abstractmethod
    def get_dependencies(self):
        pass

    @abc.abstractmethod
    def get_seeds(self, service: ConvergenceService) -> List[DatabaseSeedSpec]:
        pass


class BaseDatabaseMigration(DatabasePreparationStep):
    def __init__(self, name):
        super().__init__(name)

    @abc.abstractmethod
    def get_dependencies(self):
        pass

    @abc.abstractmethod
    def get_migration_ddl(self, formatters: SqlDialectFormatter):
        pass

    @abc.abstractmethod
    def allow_failure(self):
        pass



import threading
from datetime import datetime


class ConsoleFileDualWriter:
    def __init__(self, service, stdout):
        self.__lock = threading.Lock()
        current_file = self.get_file_path(service)
        self.current_file = current_file
        self.file = open(current_file, 'a')
        self.initialize_file_header()
        self.stdout = stdout
        self.service = service

    def write(self, line):
        self.rotate_file()
        self.file.write(line)
        self.stdout.write(line)

    def flush(self):
        self.file.flush()
        self.stdout.flush()

    def isatty(self):
        return False

    def get_file_path(self, service):
        time = str(datetime.now())
        time = time.replace('-', '').replace(':', '').replace(' ', '')[0:11] + '0'

        file_name = service.get_configuration('observability.stdout').replace('{TIME}', time)
        return service.get_configuration('observability.path') + '/' + file_name

    def rotate_file(self):
        with self.__lock:
            new_file = self.get_file_path(self.service)
            if new_file != self.current_file:
                self.current_file = new_file
                self.file.flush()
                self.file = open(self.current_file, 'a')

                self.initialize_file_header()

    def initialize_file_header(self):
        self.file.write("------------------------------------------------------------\n"
                        "Language: Python\nFormat: FastAPI\n"
                        "------------------------------------------------------------"
                        "\n\n\n")



class TableColumnBlueprint:
    def __init__(self, name, type, is_primary_key=False, is_unique=False, allow_null=False, default_value=None):
        self.name = name
        self.type = type
        self.is_primary_key = is_primary_key
        self.is_unique = is_unique
        self.allow_null = allow_null
        self.default_value = default_value


class TableBlueprint:
    def __init__(self):
        self.name = ''
        self.check_existence = False
        self.columns = []
        self.indices = []

    def add_operation_dates(self, created_at=True, updated_at=True, deleted_at=True):
        if created_at:
            self.columns.append(TableColumnBlueprint('created_at', 'timestamp', default_value='CURRENT_TIMESTAMP'))
        if updated_at:
            self.columns.append(TableColumnBlueprint('updated_at', 'timestamp', allow_null=True, default_value='NULL'))
        if deleted_at:
            self.columns.append(TableColumnBlueprint('deleted_at', 'timestamp', allow_null=True, default_value='NULL'))


class TableIndexBlueprint:
    def __init__(self, type, columns):
        self.type = type
        self.columns = columns


class TablesRelationshipBlueprint:
    def __init__(self):
        self.table = ''
        self.foreign_key = ''
        self.primary_table = ''
        self.primary_key = ''



from convergence.dto.service_status import ServiceEndpointDTO


def matches_any(internal_endpoint_urls, path):
    return path in internal_endpoint_urls


def __get_expected_authorization(authorization):
    if authorization == "@acl.allow_all()":
        return "@allow_all"
    elif authorization == "@acl.is_signed_in()":
        return "@signed_in"
    elif authorization == "@acl.not_signed_in()":
        return "@not_signed_in"
    elif authorization == "@acl.is_service()":
        return "@service_call"
    elif authorization is not None and authorization.startswith("@acl.has_authority("):
        authority = "@acl.has_authority("
        authority = authorization[len(authority) + 1:-2]
        if ')' in authority:
            return '<authorization_expression>'
        else:
            return authority
    else:
        # Python service allows expression and can't be generalized here, so allow all will pass it to the service
        # for evaluation
        return "@allow_all"


def load_service_urls(service):
    endpoints = []
    for route in service.app.routes:
        if route.include_in_schema:
            for method in route.methods:
                (info, _) = service.get_endpoint_info(route.path, method)
                ep = ServiceEndpointDTO()
                ep.url = route.path
                ep.method = method
                ep.exposed_through_gateway = info.exposed_through_gateway
                ep.expected_authorization = __get_expected_authorization(info.authorization)

                endpoints.append(ep)

    return endpoints



from convergence.dto.base_dto import ApiResponseBody


class AtomicValueDTO(ApiResponseBody):
    def __init__(self, value=None):
        self.value = value

    def get_response_body_type(self) -> str:
        return 'atomic_value'



import datetime
import time
import uuid
from typing import List

import psycopg2

from convergence.convergence_service import ConvergenceService
import convergence.database_orm as orm
from convergence.db_migrations.base_seed_migration_table import BaseCreateMigrationsTable, DatabaseMigration
from convergence.db_migrations.blueprint_formatter import SqlDialectFormatter
from convergence.db_migrations.migrations import DatabasePreparationStep, BaseDatabaseSeed, BaseDatabaseMigration
from convergence.db_migrations.postgres.postgres_relation_formatter import PostgresTableRelationBlueprintFormatter
from convergence.db_migrations.postgres.postgres_seed_formatter import PostgresDatabaseSeedFormatter
from convergence.db_migrations.postgres.postgres_table_formatter import PostgresTableFormatter
from convergence.internal.utils.dag import DirectedAcyclicGraph


class __ExternalMigrationDependency(DatabasePreparationStep):
    def __init__(self, name):
        super().__init__(name)

    def get_dependencies(self) -> List[str]:
        return []


def sort_migrations(migrations: List[DatabasePreparationStep]) \
        -> List[DatabasePreparationStep]:
    result = []
    internal_nodes = [m.name for m in migrations]
    external_dependencies = []
    for m in migrations:
        external_dependencies.extend(m.get_dependencies())

    external_dependencies = set(external_dependencies) - set(internal_nodes)
    for ex_dep in external_dependencies:
        result.append(__ExternalMigrationDependency(ex_dep))
    result.extend(migrations)

    dag = DirectedAcyclicGraph(result)
    sorted_steps = dag.topological_sort()

    result = []
    for m in sorted_steps:
        if not isinstance(m, __ExternalMigrationDependency):
            result.append(m)

    return result


def get_list_of_migrations(service: ConvergenceService, connection) -> List[str]:
    result = []

    cursor = connection.cursor()
    try:

        cursor.execute('''
            SELECT
                migration_name
            FROM database_migrations
        ''')
        records = cursor.fetchall()
        result = []

        for record in records:
            result.append(record[0])
    except:
        result = []
        connection.rollback()
    finally:
        cursor.close()

    return result


def initialize_formatters(service: ConvergenceService) -> SqlDialectFormatter:
    result = SqlDialectFormatter()
    db_type = service.get_configuration('database.type')

    if db_type == 'postgres':
        result.create_table_formatter = PostgresTableFormatter()
        result.insert_seeds_formatter = PostgresDatabaseSeedFormatter()
        result.create_relation_formatter = PostgresTableRelationBlueprintFormatter()
    else:
        raise ValueError()

    return result


def create_connection_to_db(service: ConvergenceService):
    attempts = 40
    for i in range(attempts):
        connection = __try_create_connection_to_db(service)
        if connection is not None:
            return connection
        print('Connection to DB attempt failed, waiting 0.5 second before retrying')
        time.sleep(0.5)


def __try_create_connection_to_db(service: ConvergenceService):
    try:
        return psycopg2.connect(user=service.get_configuration('database.username'),
                                password=service.get_configuration('database.password'),
                                host=service.get_configuration('database.host'),
                                port=service.get_configuration('database.port'),
                                database=service.get_configuration('database.name'))
    except:
        return None


def __execute_sql_command(command, connection):
    result = False
    error = None
    cursor = connection.cursor()

    try:
        cursor.execute(command)
        connection.commit()
        result = True
    except BaseException as ex:
        error = str(ex)
        connection.rollback()
    finally:
        cursor.close()

    return result, error


def __save_migration_state(connection, migration, command):
    entry = DatabaseMigration()
    entry.uuid = uuid.uuid4()
    entry.migration_name = migration.name
    entry.command = command
    entry.applied_timestamp = datetime.datetime.now()

    connection.add(entry)
    connection.commit()


def apply_database_migration(service, migration, connection, formatters: SqlDialectFormatter):
    command = None
    executed_with_success = False
    error = None

    if isinstance(migration, BaseDatabaseSeed):
        command = formatters.insert_seeds_formatter.to_sql(migration.get_seeds(service))
        executed_with_success = __execute_sql_command(command, connection)
    elif isinstance(migration, BaseDatabaseMigration):
        command = migration.get_migration_ddl(formatters)
        executed_with_success, error = __execute_sql_command(command, connection)
        if migration.allow_failure():
            executed_with_success = True
    else:
        return f'The migration type for {migration.name} is not supported'

    if not executed_with_success:
        return f'An error occurred while running migration {migration.name}: {error}'

    session_maker = orm.DBSessionLocal
    session = session_maker()
    __save_migration_state(session, migration, command)
    session.close()

    return None


def load_database_migrations(migrations_module: str) -> List[DatabasePreparationStep]:
    module = __get_source_module(migrations_module)
    service_migrations = __get_classes_of_module_with_base_classes(migrations_module, module, [DatabasePreparationStep],
                                                                   set())

    service_migrations.append(BaseCreateMigrationsTable())

    return service_migrations


def __is_of_base_classes(cls, base_types):
    try:
        obj = cls()
        result = False

        for t in base_types:
            result = result or isinstance(obj, t)

        return result
    except:
        return False


def __get_classes_of_module_with_base_classes(migrations_module, python_module, base_types, already_added):
    result = []
    attributes = dir(python_module)

    for attr_name in attributes:
        if attr_name.startswith('__'):
            continue

        attribute = getattr(python_module, attr_name)
        attribute_str = str(attribute)
        if attribute_str.startswith(f'<module \'{migrations_module}.'):
            result.extend(
                __get_classes_of_module_with_base_classes(migrations_module, attribute, base_types, already_added))
        if attribute_str.startswith(f'<class \'{migrations_module}.') and attribute_str not in already_added:
            already_added.add(attribute_str)
            if __is_of_base_classes(attribute, base_types):
                result.append(attribute())

    return result


def __get_source_module(migrations_module: str):
    sub_modules = migrations_module.split('.')
    result = __import__(sub_modules[0])

    for sub in sub_modules[1:]:
        result = getattr(result, sub)

    return result



import os
from typing import Dict

import yaml


def load_service_configuration(default_config_path, profile_config_path):
    default_config = __load_config_file(default_config_path)

    if profile_config_path is not None:
        profile_config = __load_config_file(profile_config_path)
        merged = __merge_config_files(default_config.copy(), profile_config)

        return __swap_environment_variables_in_config(merged)

    return __swap_environment_variables_in_config(default_config)


def __load_config_file(config_path) -> Dict:
    result = None

    with open(config_path, 'r') as yaml_config:
        result = yaml.safe_load(yaml_config)

    if result is None:
        raise ValueError(f'Unable to load config file at path: {config_path}')

    return result


def __merge_config_files(result_config: Dict, profile_config: Dict):
    # TODO: Unit tests
    for key in profile_config:
        if key in result_config:
            value = profile_config[key]
            if isinstance(value, dict):
                result_config[key] = __merge_config_files(result_config[key], value)
            else:
                result_config[key] = profile_config[key]
        else:
            result_config[key] = profile_config[key]

    return result_config


def __swap_environment_variables_in_config(config: Dict):
    swaps = {}
    for key in config:
        value = config[key]
        if isinstance(value, dict):
            __swap_environment_variables_in_config(value)
        elif isinstance(value, str):
            if value.startswith("${") and value.endswith('}'):
                variable_name = value[2:-1]
                if variable_name not in os.environ:
                    raise ValueError(f'The environment variable {variable_name} is not defined.')
                swaps[key] = os.environ[variable_name]

    for key in swaps:
        config[key] = swaps[key]

    return config



import abc


class ApiResponseHeader:
    def __init__(self):
        self.body_type: str = 'empty'
        self.status_code: int = 0
        self.code: str = ''
        self.message: str = ''
        self.request_id: str = ''
        self.parent_request_id: str = ''


class ApiResponseBody(abc.ABC):
    @abc.abstractmethod
    def get_response_body_type(self) -> str:
        pass


class ApiResponse:
    def __init__(self):
        self.header = ApiResponseHeader()
        self.body: ApiResponseBody | None = None


class FailureInfoDTO(ApiResponseBody):
    def __init__(self):
        self.status_code: int = 0
        self.code: str = ''
        self.message: str = ''

    def get_response_body_type(self) -> str:
        return 'api_failure'


def __get_list_response_body_type(body):
    result = 'empty'

    if body is not None and len(body) > 0:
        obj = body[0]
        if isinstance(obj, str):
            result = 'list[String]'
        else:
            type = obj.get_response_body_type()
            result = f'list[{type}]'

    return result


def __get_header_value(http_request, name):
    result = None
    name = name.lower()

    headers = http_request.headers.keys()
    for key in headers:
        if key.lower() == name:
            result = http_request.headers.get(key)
            break

    return result


def create_api_response(body, http_request, status_code=200) -> ApiResponse:
    if isinstance(body, ApiResponse):
        return body

    result = ApiResponse()

    if body is None:
        result.header.status_code = status_code
        result.header.body_type = 'empty'
        result.header.message = 'OK'
    elif isinstance(body, list):
        result.header.body_type = __get_list_response_body_type(body)
        result.header.status_code = status_code
        result.header.code = ''
        result.header.message = 'OK'
        result.body = body
    elif not isinstance(body, ApiResponseBody):
        raise ValueError('Api responses must be a sub class of ApiResponseBody')
    elif isinstance(body, FailureInfoDTO):
        result.header.body_type = body.get_response_body_type()
        result.header.status_code = body.status_code
        result.header.code = body.code
        result.header.message = body.message
    else:
        result.header.body_type = body.get_response_body_type()
        result.header.status_code = status_code
        result.header.code = ''
        result.header.message = 'OK'
        result.body = body

    return result



IDENTIFIER_CHARACTERS = 'qwertyuiopasdfghjklzxcvbnm1234567890QWERTYUIOPASDFGHJKLZXCVBNM_'


class AuthorizationRuleExpression:
    def __init__(self):
        self._tokens = []
        self._postfix_tokens = []
        self.rule = None

    def parse(self, rule):
        self.rule = rule
        self._tokens = self.__tokenize()
        self._postfix_tokens = self.__to_postfix()

    def evaluate(self, context):
        stack = []
        kwargs = {
            'token': context['token'],
            'req': context['req'],
            'request': context['request'],
        }

        for token in self._postfix_tokens:
            if self.__is_field(token):
                field_name = token[1:]
                if field_name not in context:
                    raise ValueError(f'Undefined field @{field_name} in authorization rule.')
                stack.append(context[field_name])
            elif token == '.':
                stack, a, b = self.__pop_top_two(stack)
                value = getattr(b, a)
                stack.append(value)
            elif token == 'and':
                stack, a, b = self.__pop_top_two(stack)
                stack.append(a and b)
            elif token == 'or':
                stack, a, b = self.__pop_top_two(stack)
                stack.append(a or b)
            elif self.__is_string(token):
                stack.append(token[1:-1])
            elif self.__is_identifier(token):
                stack.append(token)
            elif token.startswith('__call__:'):
                count = int(token[9:])
                params = []
                for i in range(count):
                    params.append(self.__peek_stack(stack))
                    stack = self.__pop_stack(stack)

                function = self.__peek_stack(stack)
                stack = self.__pop_stack(stack)

                value = function(*params, **kwargs)
                stack.append(value)
            else:
                raise ValueError(f'Unsupported token: {token}')

        if len(stack) != 1:
            raise ValueError("Invalid expression as the stack is not down to a single value!")

        return stack[0]

    def __pop_top_two(self, stack):
        a = self.__peek_stack(stack)
        stack = self.__pop_stack(stack)
        b = self.__peek_stack(stack)
        stack = self.__pop_stack(stack)

        return stack, a, b

    def __tokenize(self):
        result = []
        temp = self.rule

        while len(temp) > 0:
            ch = temp[0]
            if ch == ' ':
                temp = temp[1:]
            elif ch == '@':
                token = self.__read_field(temp)
                result.append(token)
                temp = temp[len(token):]
            elif ch == '"' or ch == '\'':
                token = self.__read_string(temp)
                result.append(token)
                temp = temp[len(token):]
            elif ch in ['.', '(', ')', ',']:
                token = ch
                result.append(token)
                temp = temp[len(token):]
            elif ch in 'qwertyuiopasdfghjklzxcvbnm1234567890QWERTYUIOPASDFGHJKLZXCVBNM_':
                token = self.__read_identifier(temp)
                result.append(token)
                temp = temp[len(token):]
            else:
                raise ValueError(f'Unable to parse expression: {self.rule}')

        return result

    def __read_field(self, temp):
        result = ['@']

        i = 1
        while i < len(temp) and temp[i] in IDENTIFIER_CHARACTERS:
            result.append(temp[i])
            i += 1

        return ''.join(result)

    def __read_string(self, temp):
        ch = temp[0]
        temp = temp[1:]

        if ch in temp:
            result = ch + temp[0:temp.index(ch)] + ch
        else:
            raise ValueError('Looks like one of the string is not properly closed.')

        return result

    def __read_identifier(self, temp):
        result = []

        i = 0
        while i < len(temp) and temp[i] in IDENTIFIER_CHARACTERS:
            result.append(temp[i])
            i += 1

        return ''.join(result)

    def __to_postfix(self):
        stack = []
        result = []
        function_call_tokens = self.__insert_function_call_operator()
        for token in function_call_tokens:
            if token.startswith('__call__:'):
                result.append(token)
            elif self.__is_operator(token):
                while self.__get_operator_precedence(token) <= self.__get_operator_precedence(self.__peek_stack(stack)):
                    result.append(self.__peek_stack(stack))
                    stack = self.__pop_stack(stack)

                stack.append(token)
            elif self.__is_string(token) or self.__is_field(token) or self.__is_identifier(token):
                result.append(token)
            elif token == ',':
                while self.__peek_stack(stack) != '(':
                    result.append(self.__peek_stack(stack))
                    stack = self.__pop_stack(stack)
            elif token == '(':
                while self.__peek_stack(stack) == '.':
                    stack = self.__pop_stack(stack)
                    result.append('.')

                stack.append('(')
            elif token == ')':
                while self.__peek_stack(stack) != '(' and self.__peek_stack(stack) is not None:
                    result.append(self.__peek_stack(stack))
                    stack = self.__pop_stack(stack)

                if self.__peek_stack(stack) == '(':
                    stack = self.__pop_stack(stack)
        while self.__peek_stack(stack) is not None:
            result.append(self.__peek_stack(stack))
            stack = self.__pop_stack(stack)

        return result

    def __is_string(self, token):
        return token.startswith('"') or token.startswith("'")

    def __is_field(self, token):
        return token.startswith('@') and self.__is_identifier(token[1:])

    def __is_identifier(self, token):
        result = True

        for ch in token:
            result = result and ch in IDENTIFIER_CHARACTERS

        return result

    def __is_operator(self, token):
        return token in ['and', 'or', '.']

    def __insert_function_call_operator(self):
        result = []
        stack = []

        depth = 0
        last_token_type = '('

        for index, token in enumerate(self._tokens):
            result.append(token)
            if self.__is_operator(token):
                last_token_type = 'operator'
            elif token == ',':
                last_token_type = 'coma'
                top_stack = self.__peek_stack(stack)
                top_stack['seen_coma'] += 1
            elif self.__is_string(token):
                last_token_type = 'string'
            elif self.__is_field(token):
                last_token_type = 'field'
            elif self.__is_identifier(token):
                last_token_type = 'identifier'
            elif token == '(':
                if last_token_type == 'identifier':
                    stack.append({
                        'depth': depth,
                        'start': index,
                        'seen_coma': 0
                    })
                last_token_type = '('
                depth += 1
                pass
            elif token == ')':
                depth -= 1
                top_stack = self.__peek_stack(stack)
                if top_stack is not None:
                    prev_depth = top_stack['depth']
                    param_count = 0
                    if index != top_stack['start'] + 1:
                        param_count = top_stack['seen_coma'] + 1

                    if prev_depth == depth:
                        result.append(f'__call__:{param_count}')
                        stack = self.__pop_stack(stack)
                last_token_type = ')'

        return result

    def __peek_stack(self, stack):
        if len(stack) > 0:
            return stack[-1]

        return None

    def __pop_stack(self, stack):
        return stack[0:-1]

    def __get_operator_precedence(self, token):
        if token is None or token == '(':
            return -1

        priorities = {
            '.': 10,
            'and': 2,
            'or': 2,
        }

        return priorities[token]



import abc


class BlueprintFormatter(abc.ABC):
    @abc.abstractmethod
    def to_sql(self, blueprint):
        pass


class SqlDialectFormatter:
    def __init__(self):
        self.create_table_formatter = None
        self.create_relation_formatter = None
        self.insert_seeds_formatter = None



import abc
import uuid
from datetime import datetime, timezone, timedelta
from enum import Enum
from typing import Any, Dict, Optional

import jwt
import requests

from convergence.constants import HEADER_REQUEST_ID, HEADER_PARENT_REQUEST_ID, HEADER_CALLER_SERVICE, \
    HEADER_CALLER_SERVICE_HASH, HEADER_CALLER_SERVICE_VERSION
from convergence.dto.bad_request_details import BadRequestDetails, BadFieldRequest
from convergence.dto.base_dto import ApiResponse, ApiResponseHeader
from convergence.internal.observability.request_log import RequestLog


class ServiceClientMode(Enum):
    INTERNAL = 1,
    EXTERNAL = 2,


class BaseServiceClient(abc.ABC):
    def __init__(self, mode: ServiceClientMode, url: Optional[str], verify_ssl: bool, type_mapper: Dict[str, Any]):
        self.__mode = mode
        self.__url = url
        self.__verify_ssl = verify_ssl
        self.__type_mapper = type_mapper

    def _make_get_call(self, request_log: RequestLog, url, required_authorization):
        headers = {
            'Content-Type': 'application/json'
        }
        self.__fill_authorization_header(required_authorization, headers)
        self.__fill_request_id_headers(request_log, headers)
        target_url = self.__get_service_url() + url

        response = requests.get(target_url, verify=self.__verify_ssl, headers=headers)
        parsed = self.__parse_json_output(response)

        return self.__build_response(parsed)

    def _make_post_call(self, request_log: RequestLog, url, required_authorization, request_body=None):
        headers = {
            'Content-Type': 'application/json'
        }
        self.__fill_authorization_header(required_authorization, headers)
        self.__fill_request_id_headers(request_log, headers)
        target_url = self.__get_service_url() + url

        content = request_body.to_dict() if request_body is not None else None
        response = requests.post(target_url, verify=self.__verify_ssl, headers=headers, json=content)
        parsed = self.__parse_json_output(response)

        return self.__build_response(parsed)

    def _make_patch_call(self, request_log: RequestLog, url, required_authorization, request_body=None):
        headers = {
            'Content-Type': 'application/json'
        }
        self.__fill_authorization_header(required_authorization, headers)
        self.__fill_request_id_headers(request_log, headers)
        target_url = self.__get_service_url() + url

        content = request_body.to_dict() if request_body is not None else None
        response = requests.patch(target_url, verify=self.__verify_ssl, headers=headers, json=content)
        parsed = self.__parse_json_output(response)

        return self.__build_response(parsed)

    def _make_put_call(self, request_log: RequestLog, url, required_authorization, request_body=None):
        headers = {
            'Content-Type': 'application/json'
        }
        self.__fill_authorization_header(required_authorization, headers)
        self.__fill_request_id_headers(request_log, headers)
        target_url = self.__get_service_url() + url

        content = request_body.to_dict() if request_body is not None else None
        response = requests.put(target_url, verify=self.__verify_ssl, headers=headers, json=content)
        parsed = self.__parse_json_output(response)

        return self.__build_response(parsed)

    def _make_delete_call(self, request_log: RequestLog, url, required_authorization, request_body=None):
        headers = {
            'Content-Type': 'application/json'
        }
        self.__fill_authorization_header(required_authorization, headers)
        self.__fill_request_id_headers(request_log, headers)
        target_url = self.__get_service_url() + url

        content = request_body.to_dict() if request_body is not None else None
        response = requests.delete(target_url, verify=self.__verify_ssl, headers=headers, json=content)
        parsed = self.__parse_json_output(response)

        return self.__build_response(parsed)

    def __build_response(self, parsed):
        response_type = parsed['header']['body_type']
        if isinstance(parsed, ApiResponse):
            return parsed
        elif response_type in self.__type_mapper:
            result = ApiResponse()
            result.header = self.__parse_header(parsed['header'])
            result.body = self.__type_mapper[response_type].from_dict(parsed['body'])
            return result
        elif response_type in ['empty', 'api_failure']:
            result = ApiResponse()
            result.header = self.__parse_header(parsed['header'])
            return result
        elif response_type == 'request_error_info':
            result = ApiResponse()
            result.header = self.__parse_header(parsed['header'])
            result.body = self.__parse_bad_request_error_info(parsed['body'])
            return result
        else:
            raise ValueError('Unexpected error occurred while parsing the API response.')

    def __fill_authorization_header(self, authorization, headers):
        if authorization in ["@allow_all", "@not_signed_in"]:
            # No need to include an authorization header
            pass
        elif authorization == "@signed_in":
            jwt_header = f'Bearer {self.__create_jwt(None, False)}'
            headers['Authorization'] = jwt_header
        elif authorization == "@service_call":
            jwt_header = f'Bearer {self.__create_jwt(None, True)}'
            headers['Authorization'] = jwt_header
        else:
            jwt_header = f'Bearer {self.__create_jwt(authorization, False)}'
            headers['Authorization'] = jwt_header

    def __create_jwt(self, authorities, service_jwt):
        signing_key = self.get_jwt_signing_key()
        info = self.get_service_info()

        payload = {
            "sub": info.SERVICE_NAME,
            "iss": info.SERVICE_NAME,
            "exp": datetime.now(tz=timezone.utc) + timedelta(minutes=1),
        }

        if service_jwt:
            payload['is_inter_service_call'] = True

        if authorities is not None:
            authorities = [] if authorities is None else authorities
            authorities = [authorities] if not isinstance(authorities, list) else authorities
            payload['authorities'] = authorities

        return jwt.encode(payload, signing_key, algorithm="ES512")

    def __get_service_url(self):
        if self.__url is None:
            raise ValueError('To be implemented: get the service URL from core-service')

        return self.__url

    def __parse_json_output(self, response):
        try:
            return response.json()
        except:  # noqa
            result = ApiResponse()

            result.header.body_type = 'api_failure'
            result.header.status_code = response.status_code
            result.header.code = 'err_unable_to_parse_service_output'
            result.header.message = 'The response got from the service is not valid JSON and can not be parsed.'
            result.header.request_id = None
            result.header.parent_request_id = None

            return result

    def __parse_header(self, parsed):
        result = ApiResponseHeader()

        result.body_type = parsed['body_type']
        result.status_code = parsed['status_code']
        result.code = parsed['code']
        result.message = parsed['message']
        result.request_id = parsed['request_id']
        result.parent_request_id = parsed['parent_request_id']

        return result

    def __fill_request_id_headers(self, request_log, headers):
        info = self.get_service_info()

        if self.__mode == ServiceClientMode.INTERNAL:
            headers[HEADER_REQUEST_ID] = str(uuid.uuid4())
            headers[HEADER_PARENT_REQUEST_ID] = request_log.request_identifier
            headers[HEADER_CALLER_SERVICE] = info.SERVICE_NAME
            headers[HEADER_CALLER_SERVICE_HASH] = info.SERVICE_VERSION_HASH
            headers[HEADER_CALLER_SERVICE_VERSION] = info.SERVICE_VERSION

    @abc.abstractmethod
    def get_jwt_signing_key(self):
        pass

    @abc.abstractmethod
    def get_service_info(self):
        pass

    def __parse_bad_request_error_info(self, data):
        result = BadRequestDetails()

        for e in data['errors']:
            detail = BadFieldRequest()
            detail.location = e['location']
            detail.field = e['field']
            detail.error_messages = e['error_messages']

            result.errors.append(detail)

        return result



import base64
import json
import re
import uuid
from io import StringIO

import yaml
from pydantic import ValidationError
from pydantic_core import InitErrorDetails, PydanticCustomError
from requests import PreparedRequest

from convergence.dto.bad_request_details import BadRequestDetails, BadFieldRequest
from convergence.dto.base_dto import ApiResponse
from convergence.helpers.errors import INVALID_DATA
from convergence.validations.input_validation_filter import create_map_id


class PydanticErrorInfo():
    def __init__(self, field=''):
        self.field = field
        self.__error_list = []

    def __add_error(self, message, field_name, value):
        self.__error_list.append((message, field_name, value))

    def has_errors(self):
        return len(self.__error_list) > 0

    def raise_error(self):
        if len(self.__error_list) == 0:
            return

        validation_errors: list[InitErrorDetails] = list()
        for (message, field_name, value) in self.__error_list:
            validation_errors.append(
                InitErrorDetails(
                    type=PydanticCustomError("static_validation_check", message),
                    loc=(),
                    input=value,
                    ctx={},
                )
            )

        raise ValidationError.from_exception_data(title='static validation error', line_errors=validation_errors)

    def build_bad_request_response_for_invalid_path_variable(self, request_log):
        result = ApiResponse()

        result.body = BadRequestDetails()
        result.header.status_code = 400
        result.header.code = INVALID_DATA
        result.header.message = 'The request input is invalid, refer to body for details.'
        result.header.request_id = request_log.request_identifier
        result.header.parent_request_id = request_log.parent_request_identifier
        result.header.body_type = result.body.get_response_body_type()

        errors = {}
        for (message, field_name, value) in self.__error_list:
            map_id = create_map_id(field_name)
            error_entry = None
            if map_id in errors:
                error_entry = errors[map_id]
            else:
                error_entry = BadFieldRequest()
                error_entry.location = 'url'
                error_entry.field = field_name
                result.body.errors.append(error_entry)
                errors[map_id] = error_entry

            error_entry.error_messages.append(message)

        return result

    def verify_alpha(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Alpha validation is only for strings', value)
        else:
            allowed_chars = 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'
            for ch in value:
                if ch not in allowed_chars:
                    self.__add_error("Failing to pass validation: 'alpha'", field_name, value)
                    break

    def verify_alphanum(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Alphanum validation is only for strings', value)
        else:
            allowed_chars = 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890'
            for ch in value:
                if ch not in allowed_chars:
                    self.__add_error("Failing to pass validation: 'alphanum'", field_name, value)
                    break

    def verify_min(self, field_name, value, constraint):
        if not isinstance(value, int) and not isinstance(value, float):
            self.__add_error('`min` validation is only for numeric values', value)
        elif value < constraint:
            self.__add_error("Failing to pass validation: 'min'", field_name, value)

    def verify_max(self, field_name, value, constraint):
        if not isinstance(value, int) and not isinstance(value, float):
            self.__add_error('`max` validation is only for numeric values', value)
        elif value > constraint:
            self.__add_error("Failing to pass validation: 'max'", field_name, value)

    def verify_oneof(self, field_name, value, allowed_values):
        if not isinstance(value, str):
            self.__add_error('`oneof` validation is only for string values', value)
        elif value not in allowed_values:
            self.__add_error("Failing to pass validation: 'oneof'", field_name, value)

    def verify_min_length(self, field_name, value, constraint):
        if not isinstance(value, str) and not isinstance(value, list):
            self.__add_error('`min_length` validation is only for strings and lists.', value)
        elif len(value) < constraint:
            self.__add_error("Failing to pass validation: 'min_length'", field_name, value)

    def verify_max_length(self, field_name, value, constraint):
        if not isinstance(value, str) and not isinstance(value, list):
            self.__add_error('`max_length` validation is only for strings and lists.', value)
        elif len(value) > constraint:
            self.__add_error("Failing to pass validation: 'max_length'", field_name, value)

    def verify_base64(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Base64 validation is only for strings', value)
        else:
            if not self.__is_base_64(value):
                self.__add_error("Failing to pass validation: 'base64'", field_name, value)

    def __is_base_64(self, value):
        try:
            base64.b64decode(value)
            return True
        except:  # noqa
            return False

    def __is_safe_url_base_64(self, value):
        try:
            base64.urlsafe_b64decode(value)
            return True
        except:  # noqa
            return False

    def verify_email(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Email validation is only for strings', value)
        else:
            regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'
            if not re.fullmatch(regex, value):
                self.__add_error("Failing to pass validation: 'email'", field_name, value)

    def verify_url(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('URL validation is only for strings', value)
        else:
            prepared_request = PreparedRequest()
            try:
                prepared_request.prepare_url(value, None)
                return prepared_request.url
            except:  # noqa
                self.__add_error("Failing to pass validation: 'url'", field_name, value)

    def verify_uuid(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('UUID validation is only for strings', value)
        else:
            try:
                uuid.UUID(value)
            except:  # noqa
                self.__add_error("Failing to pass validation: 'uuid'", field_name, value)

    def verify_contain(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Contain validation is only for strings', value)
        else:
            if pattern not in value:
                self.__add_error("Failing to pass validation: 'contain'", field_name, value)

    def verify_not_contain(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Not contain validation is only for strings', value)
        else:
            if pattern in value:
                self.__add_error("Failing to pass validation: 'not_contain'", field_name, value)

    def verify_string_start_with(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Starts with validation is only for strings', value)
        else:
            if not value.startswith(pattern):
                self.__add_error("Failing to pass validation: 'string_start_with'", field_name, value)

    def verify_string_not_start_with(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Not starts with validation is only for strings', value)
        else:
            if value.startswith(pattern):
                self.__add_error("Failing to pass validation: 'string_not_start_with'", field_name, value)

    def verify_string_end_with(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Ends with validation is only for strings', value)
        else:
            if not value.endswith(pattern):
                self.__add_error("Failing to pass validation: 'string_end_with'", field_name, value)

    def verify_string_not_end_with(self, field_name, allow_none, value, pattern):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Not ends with validation is only for strings', value)
        else:
            if value.endswith(pattern):
                self.__add_error("Failing to pass validation: 'string_not_end_with'", field_name, value)

    def verify_yaml(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('Not ends with validation is only for strings', value)
        else:
            a = yaml.safe_load(StringIO(value))
            if not isinstance(a, dict):
                self.__add_error("Failing to pass validation: 'yaml'", field_name, value)

    def verify_jwt(self, field_name, allow_none, value):
        if allow_none and value is None:
            return

        if value is None:
            self.__add_error('Should not be null', value)
        elif not isinstance(value, str):
            self.__add_error('JWT validation is only for strings', value)
        else:
            comps = value.split('.')
            if len(comps) != 3:
                self.__add_error("Failing to pass validation: 'jwt'", field_name, value)
            elif not self.__is_safe_url_base_64(comps[0]) or not self.__is_safe_url_base_64(comps[1]):
                self.__add_error("Failing to pass validation: 'jwt'", field_name, value)
            else:
                p1 = base64.urlsafe_b64decode(comps[0]).decode('utf-8')
                p2 = base64.urlsafe_b64decode(comps[1]).decode('utf-8')

                try:
                    json.loads(p1)
                    json.loads(p2)
                except:  # noqa
                    self.__add_error("Failing to pass validation: 'jwt'", field_name, value)



API_NOT_IMPLEMENTED = "err_api_not_implemented"
API_VERSION_NOT_FOUND = "err_api_version_not_found"
API_INTERNAL_ERROR = "err_api_internal_error"
API_MISSING_REQUEST_ID = "err_api_missing_request_id"
EXPIRED_AUTHORIZATION_TOKEN = "err_authorization_token_expired"
USER_BLOCKED = "err_user_blocked"
API_RESOURCE_NOT_FOUND = "err_api_resource_not_found"
SERVICE_NOT_FOUND = "err_api_service_not_found"
API_RESOURCE_ALREADY_EXISTS = "err_api_resource_already_exists"
API_INVALID_ENTITY_STATE = "err_api_invalid_entity_state"
INVALID_DATA = "err_api_invalid_data"
API_UNPARSABLE_JSON = "err_unparseable_input"
INVALID_AUTHENTICATION_CREDENTIALS = "err_auth_invalid_credentials"
UNHANDLED_EXCEPTION = "err_unknown_exception"
INVALID_AUTHENTICATION_STATE = "err_auth_invalid_state"
ERR_ACCESS_DENIED = "err_acl_access_denied"
EMPTY_QUEUE = "wrn_nothing_queued_for_execution"
API_STORAGE_NONE_EMPTY_FOLDER = "err_storage_folder_none_empty"
API_SERVICE_DOWN_ERROR = "err_external_service_down"
API_METHOD_NOT_ALLOWED = "err_method_not_allowed"



from convergence.internal.services.base_internal_service import BaseExternalService


class AuthenticationMicroService(BaseExternalService):
    def __init__(self, service, url):
        super().__init__(service, url, service_jwt=True)

    def register_service_authority(self, request):
        return self.post_request('/authentication/internal-services/register-authority', request)



VERSION = '1.0.7'
VERSION_HASH = 'cc653354eb327c37adc66dfd11163ea0c919e00c'
BUILD_DATE = '2024-01-29 12:28:51 UTC'

HEADER_REQUEST_ID = "X-CONVERGENCE-REQUEST-ID"
HEADER_CALLER_SERVICE = "X-CONVERGENCE-CALLER-SERVICE"
HEADER_CALLER_SERVICE_HASH = "X-CONVERGENCE-CALLER-SERVICE-HASH"
HEADER_CALLER_SERVICE_VERSION = "X-CONVERGENCE-CALLER-SERVICE-VERSION"
HEADER_PARENT_REQUEST_ID = "X-CONVERGENCE-PARENT-REQUEST-ID"
